/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include <boost/lexical_cast.hpp>
#include <common/base/Date.h>
#include <fstream>
#include <iostream>
#include "RemoteFilePub.h"

namespace test {

using namespace std;
using namespace common;
using namespace boost;
using namespace cameo;

const std::string RemoteFilePub::TYPE = "remote_file_pub";
const std::string RemoteFilePub::REMOTE_APPLICATION = "nfilepub";
const std::string RemoteFilePub::FILE_PUBLISHER = "file_publisher";
const std::string RemoteFilePub::RESULT_PUBLISHER = "file_result";

RemoteFilePub::RemoteFilePub(const std::string& name) :
	ExperimentController(name),
	controller::Stoppable(this) {

	serverEndpoint.init(this, SAVE, "cameo_server");
	initialized.init(this, NOSAVE, "initialized");
	filename.init(this, SAVE, "filename");
}

RemoteFilePub::RemoteFilePub(const RemoteFilePub& controller) :
	ExperimentController(controller),
	controller::Stoppable(this) {
}

RemoteFilePub::~RemoteFilePub() {

	// Stop the subscriber
	m_subscriber->cancel();

	// Notify the application
	m_publisher->sendEnd();

	// Wait for the termination
	application::State state = m_remoteApplication->waitFor();

	cout << "Remote application " << m_remoteApplication->getName() << " terminated with state " << application::toString(state) << endl;
}

void RemoteFilePub::postConfiguration() {

	if (serverEndpoint() == "") {
		m_server.reset(new Server(application::This::getServer().getEndpoint()));

	} else {
		m_server.reset(new Server(serverEndpoint()));
	}

	initialized = initApplication();
}

bool RemoteFilePub::initApplication() {

	// Do not initialize if it is already done
	if (initialized()) {
		return true;
	}

	// Start the remote server
	if (!m_server->isAvailable(1000)) {
		cout << "Remote server is not available" << endl;
		return false;
	}

	cout << "Remote server is connected" << endl;

	m_remoteApplication = m_server->connect(REMOTE_APPLICATION);
	if (m_remoteApplication->exists()) {
		// The application exists from a previous server session
		m_remoteApplication->kill();
		application::State state = m_remoteApplication->waitFor();
		cout << "Terminated remote application with state " << application::toString(state) << endl;
	}

	m_remoteApplication = m_server->start(REMOTE_APPLICATION);

	if (!m_remoteApplication->exists()) {
		cout << "No remote application" << endl;
		return false;
	}

	// Create the publisher after the start of the application, because it is a blocking call
	m_publisher = application::Publisher::create(FILE_PUBLISHER, 1);
	cout << "Publisher " << *m_publisher << endl;

	// Wait for the subscriber
	bool sync = m_publisher->waitForSubscribers();

	// Create the subscriber after the publisher
	m_subscriber = application::Subscriber::create(*m_remoteApplication, RESULT_PUBLISHER);
	cout << "subscriber " << *m_subscriber << endl;

	// Application initialized
	initialized = true;

	return true;
}

void RemoteFilePub::sendFile() {

	if (!initApplication()) {
		return;
	}

	ifstream numorFile;
	numorFile.open(filename().c_str());

	stringstream stream;
	stream << numorFile.rdbuf();

	Date begin;

	string content = stream.str();

	// Publish the message
	m_publisher->sendBinary(content);

	// Wait for result
	vector<int32> result;
	if (m_subscriber->receive(result)) {
		Date end;
		double ms = (end - begin).getMilliseconds();
		cout << "Sent and received file in " << ms << " ms (" << (content.size() / ms) << " ko/s)" << endl;
		if (content.size() != result[0]) {
			cout << "\tProblem with size" << endl;
		}
	}
}

void RemoteFilePub::start() {

	while (true) {
		// Send the numor
		sendFile();

		if (isStopped()) {
			break;
		}
	}
}

void RemoteFilePub::stop() {

}

}
